# Copyright 2022 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################


from mio import cfg
from mio import clean
from mio import cov
from mio import dox
from mio import results
from mio import cache
from mio import common
from mio import eal
from mio import install
from alive_progress import alive_bar
from multiprocessing import Process
import time
import os
from datetime import datetime
import yaml
from yaml.loader import SafeLoader
import math


regex_define_pattern  = "\+define\+((?:\w|_|\d)+)(?:\=((?:\w|_|\d)+))?"
regex_plusarg_pattern = "\+((?:\w|_|\d)+)(?:\=((?:\w|_|\d)+))?"
bar = None
seconds_waited = 0


def main(cli_args):
    ip_name = cfg.target_ip_name
    
    if not eal.check_simulator_bin_path(cli_args):
        common.fatal("Simulator '" + cli_args['simulator'] + "' not installed properly or environment variable missing")
    
    defines   = convert_cli_args_to_defines()
    plus_args = convert_cli_args_to_plusargs()
    create_sim_directories()
    
    if cli_args['cmp']:
        if cfg.target_ip['installed'] == False:
            install_deps_str = ""
            num_deps = len(cfg.target_ip['data']['dependencies'])
            while (install_deps_str != "y") and (install_deps_str != "n"):
                install_deps_str = common.prompt(str(num_deps) + " IP dependencies must first be installed.  Would you like to do so now? [y/n]").strip().lower()
            if install_deps_str == "y":
                local_str = ""
                while (local_str != "y") and (local_str != "n"):
                    local_str = common.prompt("Local install (vs. global)? [y/n]").strip().lower()
                if local_str == "y":
                    global_install = False
                else:
                    global_install = True
                common.info("Installing " + str(num_deps) + " dependencies ...")
                try:
                    install.install_all_ips_for_target(global_install)
                except Exception as e:
                    common.fatal("Error during '" + ip_name + "' IP installation: " + str(e))
            else:
                common.fatal("Cannot continue without first installing '" + ip_name + "' IP dependencies.")
    
    check_dependencies()
    
    if cli_args['cmp']:
        common.banner("Compiling IP '" + ip_name + "'")
        num_deps = len(cfg.target_ip['data']['dependencies'])
        for dep in cfg.target_ip['data']['dependencies']:
            if dep['name'] not in cfg.ip_local_cache:
                dep_ip = cache.get_ip(dep['name'])
                if dep_ip['compiled'][cli_args['simulator']]:
                    num_deps -= 1
        if num_deps > 0:
            common.info("Compiling " + str(num_deps) + " dependencies")
            cmp_dependencies(num_deps, defines, cli_args)
        if 'dut' in cfg.target_ip['data']:
            dut_ip_name = cfg.target_ip['data']['dut']['name']
            if dut_ip_name in cfg.job_history:
                est_time = 0
                curr_est_time = 0
                if 'compilation' in cfg.job_history[ip_name]:
                    for job in cfg.job_history[ip_name]['compilation']:
                        start = datetime.strptime(job['timestamp_start'], "%Y/%m/%d-%H:%M:%S")
                        end   = datetime.strptime(job['timestamp_end'  ], "%Y/%m/%d-%H:%M:%S")
                        curr_est_time = end - start
                        curr_est_time = divmod(curr_est_time.seconds, 60)[1]
                        est_time += curr_est_time
                    est_time = math.ceil(est_time / len(cfg.job_history[ip_name]))
                    if est_time > 0:
                        p = Process(target=progress_bar, args=(est_time,))
                        p.start()
                        cmp_dut(defines, cli_args)
                        if bar != None:
                            bar(est_time - seconds_waited)
                        p.join()
                    else:
                        cmp_dut(defines, cli_args)
                else:
                    cmp_dut(defines, cli_args)
            else:
                cmp_dut(defines, cli_args)
        
        if ip_name in cfg.job_history:
            if 'compilation' in cfg.job_history[ip_name]:
                est_time = 0
                curr_est_time = 0
                for job in cfg.job_history[ip_name]['compilation']:
                    start = datetime.strptime(job['timestamp_start'], "%Y/%m/%d-%H:%M:%S")
                    end   = datetime.strptime(job['timestamp_end'  ], "%Y/%m/%d-%H:%M:%S")
                    curr_est_time = end - start
                    curr_est_time = divmod(curr_est_time.seconds, 60)[1]
                    est_time += curr_est_time
                est_time = math.ceil(est_time / len(cfg.job_history[ip_name]))
                if est_time > 0:
                    p = Process(target=progress_bar, args=(est_time,))
                    p.start()
                    cmp_target_ip(defines, cli_args)
                    if bar != None:
                        bar(est_time - seconds_waited)
                    p.join()
                else:
                    cmp_target_ip(defines, cli_args)
            else:
                cmp_target_ip(defines, cli_args)
        else:
            cmp_target_ip(defines, cli_args)
    
    if cli_args['elab']:
        common.banner("Elaborating IP '" + ip_name + "'")
        est_time = 0
        curr_est_time = 0
        if ip_name in cfg.job_history:
            if 'elaboration' in cfg.job_history[ip_name]:
                for job in cfg.job_history[ip_name]['elaboration']:
                    start = datetime.strptime(job['timestamp_start'], "%Y/%m/%d-%H:%M:%S")
                    end   = datetime.strptime(job['timestamp_end'  ], "%Y/%m/%d-%H:%M:%S")
                    curr_est_time = end - start
                    curr_est_time = divmod(curr_est_time.seconds, 60)[1]
                    est_time += curr_est_time
                est_time = math.ceil(est_time / len(cfg.job_history[ip_name]))
                if est_time > 0:
                    p = Process(target=progress_bar, args=(est_time,))
                    p.start()
                    eal.elaborate(cfg.target_ip, defines, cli_args)
                    if bar != None:
                        bar(est_time - seconds_waited)
                    p.join()
                else:
                    eal.elaborate(cfg.target_ip, defines, cli_args)
            else:
                eal.elaborate(cfg.target_ip, defines, cli_args)
        else:
            eal.elaborate(cfg.target_ip, defines, cli_args)
    else:
        if cli_args['sim'] == False:
            print_end_of_compilation_message(cli_args)
    
    if cli_args['sim']:
        common.banner("Simulating IP '" + ip_name + "'")
        eal.simulate(cfg.target_ip, plus_args, cli_args, is_regression=False)
        if cli_args['cmp'] == True:
            print_end_of_compilation_message(cli_args)
        if cli_args['elab'] == True:
            print_end_of_elaboration_message(cli_args)
        print_end_of_simulation_message(cli_args)
    else:
        if cli_args['elab'] == True:
            if cli_args['cmp'] == True:
                print_end_of_compilation_message(cli_args)
            print_end_of_elaboration_message(cli_args)


def progress_bar(seconds):
    global bar
    global seconds_waited
    tenth_seconds_waited = 0
    with alive_bar(seconds*10, bar = 'smooth', stats="{eta} estimated", monitor=False, elapsed=True) as bar:
        bar.text("estimated")
        for x in range(seconds*10):
            time.sleep(0.1)
            tenth_seconds_waited += 1
            if (tenth_seconds_waited == 10):
                tenth_seconds_waited = 0
                seconds_waited += 1
            bar()


def convert_cli_args_to_defines():
    defines = {}
    if cfg.cli_args.add_args != None:
        all_args = re.sub("\"", "", cfg.cli_args.add_args)
        for arg in all_args:
            result = re.match(regex_define_pattern, arg)
            if result:
                define_name = result.group(1)
                if len(results.group) > 1:
                    define_value = result.group(2)
                else:
                    define_value = ""
                defines[define_name] = define_value
                common.dbg("Added define '" + define_name + "' with value '" + define_value + "' to list")
    return defines


def convert_cli_args_to_plusargs():
    plus_args = {}
    if cfg.cli_args.add_args != None:
        all_args = re.sub("\"", "", cfg.cli_args.add_args)
        for arg in all_args:
            result = re.match(regex_define_pattern, arg)
            if not result:
                result = re.match(regex_plusarg_pattern, arg)
                if result:
                    arg_name = result.group(1)
                    if len(results.group) > 1:
                        arg_value = result.group(2)
                    else:
                        arg_value = ""
                    plus_args[arg_name] = arg_value
                    common.dbg("Added plus arg '" + arg_name + "' with value '" + arg_value + "' to list")
    return plus_args


def create_sim_directories():
    common.dbg("Creating sim directories")
    common.create_dir(cfg.sim_dir)
    common.create_dir(cfg.sim_output_dir)
    common.create_dir(cfg.sim_output_dir + "/vivado"          )
    common.create_dir(cfg.sim_output_dir + "/vivado/sim"      )
    common.create_dir(cfg.sim_output_dir + "/vivado/local"    )
    common.create_dir(cfg.sim_output_dir + "/vivado/external" )
    common.create_dir(cfg.sim_output_dir + "/metrics"         )
    common.create_dir(cfg.sim_output_dir + "/metrics/sim"     )
    common.create_dir(cfg.sim_output_dir + "/metrics/local"   )
    common.create_dir(cfg.sim_output_dir + "/metrics/external")
    common.create_dir(cfg.sim_output_dir + "/vcs"             )
    common.create_dir(cfg.sim_output_dir + "/vcs/sim"         )
    common.create_dir(cfg.sim_output_dir + "/vcs/local"       )
    common.create_dir(cfg.sim_output_dir + "/vcs/external"    )
    common.create_dir(cfg.sim_output_dir + "/xcelium"         )
    common.create_dir(cfg.sim_output_dir + "/xcelium/sim"     )
    common.create_dir(cfg.sim_output_dir + "/xcelium/local"   )
    common.create_dir(cfg.sim_output_dir + "/xcelium/external")
    common.create_dir(cfg.sim_output_dir + "/questa"          )
    common.create_dir(cfg.sim_output_dir + "/questa/sim"      )
    common.create_dir(cfg.sim_output_dir + "/questa/local"    )
    common.create_dir(cfg.sim_output_dir + "/questa/external" )
    common.create_dir(cfg.sim_output_dir + "/riviera"         )
    common.create_dir(cfg.sim_output_dir + "/riviera/sim"     )
    common.create_dir(cfg.sim_output_dir + "/riviera/local"   )
    common.create_dir(cfg.sim_output_dir + "/riviera/external")
    common.create_dir(cfg.sim_dir + "/cmp")
    common.create_dir(cfg.sim_dir + "/elab")
    common.create_dir(cfg.sim_results_dir)


def check_dependencies():
    for dep in cfg.target_ip['data']['dependencies']:
        found_ip = False
        dep_name = dep['name']
        if dep_name in cfg.ip_local_cache:
            found_ip = True
        else:
            for vendor in cfg.ip_external_cache:
                if dep_name in cfg.ip_external_cache[vendor]:
                    found_ip = True
        if not found_ip:
            common.fatal("Could not find IP dependency '" + dep_name + "'")


def cmp_dependencies(num_deps, defines, cli_args):
    with alive_bar(num_deps-1, bar = 'smooth') as bar:
        for dep in cfg.target_ip['data']['dependencies']:
            dep_name = dep['name']
            if dep_name == "uvm":
                continue
            if dep_name in cfg.ip_local_cache:
                dep_ip = cfg.ip_local_cache[dep_name]
                common.dbg("Compiling IP dependency '" + dep_name + "'")
                bar.text(dep_name)
                eal.compile_ip(dep_ip, defines, cli_args, local=True)
                bar()
            else:
                dep_ip = cache.get_external_ip(dep_name)
                if not dep_ip['compiled'][cli_args['simulator']]:
                    common.dbg("Compiling IP dependency '" + dep_name + "'")
                    bar.text(dep_name)
                    eal.compile_ip(dep_ip, defines, cli_args, local=False)
                    bar()


def cmp_dut(defines, cli_args):
    found_dut_ip = False
    dut_type = cfg.target_ip['data']['dut']['ip-type']
    dut_ip_name = ""
    if dut_type == "fsoc":
        dut_ip_name = cfg.target_ip['data']['dut']['full-name'].strip()
        fsoc_target = cfg.target_ip['data']['dut']['target'].strip()
        if dut_ip_name in cfg.fsoc_cache:
            found_dut_ip = True
            common.dbg("Found FuseSoC DUT '" + dut_ip_name + "'")
            core = cfg.fsoc_cache[dut_ip_name]
            ip_path = core['ip_path'].strip()
            if not core['installed'] or cfg.cli_args.F:
                flist_path = eal.invoke_fsoc(cfg.target_ip, core, dut_ip_name, ip_path, fsoc_target, cli_args)
                core['installed'] = True
                eal.compile_fsoc_core(flist_path, core, defines, cli_args)
            else:
                common.info("Skipping installation of DUT FuseSoC core '" + dut_ip_name + "'.")
    elif dut_type == "mio":
        local_ip = False
        dut_ip_name = cfg.target_ip['data']['dut']['name']
        if dut_ip_name in cfg.ip_local_cache:
            local_ip = True
            found_dut_ip = True
            dut_ip = cfg.ip_local_cache[dut_ip_name]
            common.dbg("Found local IP DUT '" + dut_ip_name + "'")
        else:
            dut_ip = get_external_ip(dut_ip_name)
            found_dut_ip = True
            common.dbg("Found external IP DUT '" + dut_ip_name + "'")
        if not dut_ip['compiled'][cli_args['simulator']]:
            dut_sub_type = dut_ip['data']['ip']['sub-type'].lower().strip()
            if dut_sub_type == "vivado":
                common.info("Compiling Vivado Project DUT IP '" + dut_ip_name + "'")
                eal.compile_vivado_project(dut_ip, defines, cli_args, local=local_ip)
            elif dut_sub_type == "flist":
                common.info("Compiling DUT IP %s ...")
                eal.compile_ip(dut_ip, defines, cli_args, local=local_ip)
            else:
                common.fatal("Invalid DUT sub-type '" + dut_sub_type + "'")
        else:
            common.info("Skipping compilation of DUT IP '" + dut_ip_name + "'.")
    else:
        common.fatal("DUT type '" + dut_type + "' not supported.")
    if found_dut_ip == False:
        common.fatal("Could not find DUT '" + dut_ip_name + "'.")


def cmp_target_ip(defines, cli_args):
    ip_name = cfg.target_ip_name
    if ip_name in cfg.ip_local_cache:
        ip_yml = cfg.ip_local_cache[ip_name]
        local_ip = True
    else:
        ip_yml = get_external_ip(ip_name)
        local_ip = False
    if 'xilinx-libs' in ip_yml['data']['hdl-src']:
        common.info("Compiling '" + ip_name + "'")
        eal.compile_vivado_project(ip_yml, defines, cli_args, local=local_ip)
    else:
        common.info(f"Compiling '{ip_name}'")
        eal.compile_ip(ip_yml, defines, cli_args, local=local_ip)



def print_end_of_compilation_message(cli_args):
    common.info("************************************************************************************************************************")
    common.info("  Compilation results")
    common.info("************************************************************************************************************************")
    common.info("  emacs " + cfg.sim_dir + "/cmp/" + cfg.target_ip_name + "." + common.get_simulator_short_name(cli_args) + ".cmp.log &")
    common.info("  gvim  " + cfg.sim_dir + "/cmp/" + cfg.target_ip_name + "." + common.get_simulator_short_name(cli_args) + ".cmp.log &")
    common.info("  vim   " + cfg.sim_dir + "/cmp/" + cfg.target_ip_name + "." + common.get_simulator_short_name(cli_args) + ".cmp.log")
    common.info("")


def print_end_of_elaboration_message(cli_args):
    common.info("************************************************************************************************************************")
    common.info("  Elaboration results")
    common.info("************************************************************************************************************************")
    common.info("  emacs " + cfg.sim_dir + "/elab/" + cfg.target_ip_name + "." + common.get_simulator_short_name(cli_args) + ".elab.log &")
    common.info("  gvim  " + cfg.sim_dir + "/elab/" + cfg.target_ip_name + "." + common.get_simulator_short_name(cli_args) + ".elab.log &")
    common.info("  vim   " + cfg.sim_dir + "/elab/" + cfg.target_ip_name + "." + common.get_simulator_short_name(cli_args) + ".elab.log")
    common.info("")


def print_end_of_simulation_message(cli_args):
    results_path = cli_args['results_path']
    common.info("************************************************************************************************************************")
    common.info("  Simulation results")
    common.info("************************************************************************************************************************")
    if (cli_args['waves'] == True):
        if cli_args['simulator'] == "vivado":
            common.info("  Waveforms: $VIVADO_HOME/xsim -gui " + results_path + "/waves.wdb &")
        elif cli_args['simulator'] == "vcs":
            common.info("  Waveforms: $VCS_HOME/dve -gui " + results_path + "/waves.wdb &")
        elif cli_args['simulator'] == "metrics":
            common.info("  Waveforms: $METRICS_HOME/??? -gui " + results_path + "/waves.wdb &")
        elif cli_args['simulator'] == "xcelium":
            common.info("  Waveforms: $XCELIUM_HOME/simvision -gui " + results_path + "/waves.wdb &")
        elif cli_args['simulator'] == "questa":
            common.info("  Waveforms: $QUESTA_HOME/visualizer -gui " + results_path + "/waves.wdb &")
        elif cli_args['simulator'] == "riviera":
            common.info("  Waveforms: $RIVIERA_HOME/??? -gui " + results_path + "/waves.wdb &")
    common.info("")
    common.info("  Main log: emacs " + results_path + "/sim.log &")
    common.info("            gvim  " + results_path + "/sim.log &")
    common.info("            vim   " + results_path + "/sim.log")
    common.info("")
    common.info("  Transaction logs: pushd " + results_path + "/trn_log")
    common.info("  Test result dir : pushd " + results_path)
    common.info("")


